/*
 * SPDX-FileCopyrightText: syuilo and misskey-project
 * SPDX-License-Identifier: AGPL-3.0-only
 */

import { Brackets } from 'typeorm';
import { Inject, Injectable } from '@nestjs/common';
import type { NotesRepository } from '@/models/_.js';
import { Endpoint } from '@/server/api/endpoint-base.js';
import { NoteEntityService } from '@/core/entities/NoteEntityService.js';
import { DI } from '@/di-symbols.js';
import { CacheService } from '@/core/CacheService.js';
import { IdService } from '@/core/IdService.js';
import { QueryService } from '@/core/QueryService.js';
import { MetaService } from '@/core/MetaService.js';
import { MiLocalUser } from '@/models/User.js';
import { FanoutTimelineEndpointService } from '@/core/FanoutTimelineEndpointService.js';
import { FanoutTimelineName } from '@/core/FanoutTimelineService.js';
import { ApiError } from '@/server/api/error.js';
import { isQuote, isRenote } from '@/misc/is-renote.js';

export const meta = {
	tags: ['users', 'notes'],

	res: {
		type: 'array',
		optional: false, nullable: false,
		items: {
			type: 'object',
			optional: false, nullable: false,
			ref: 'Note',
		},
	},

	errors: {
		noSuchUser: {
			message: 'No such user.',
			code: 'NO_SUCH_USER',
			id: '27e494ba-2ac2-48e8-893b-10d4d8c2387b',
		},

		bothWithRepliesAndWithFiles: {
			message: 'Specifying both withReplies and withFiles is not supported',
			code: 'BOTH_WITH_REPLIES_AND_WITH_FILES',
			id: '91c8cb9f-36ed-46e7-9ca2-7df96ed6e222',
		},
	},
} as const;

export const paramDef = {
	type: 'object',
	properties: {
		userId: { type: 'string', format: 'misskey:id' },
		withReplies: { type: 'boolean', default: false },
		withRepliesToSelf: { type: 'boolean', default: true },
		withQuotes: { type: 'boolean', default: true },
		withRenotes: { type: 'boolean', default: true },
		withNonPublic: { type: 'boolean', default: true },
		withChannelNotes: { type: 'boolean', default: false },
		limit: { type: 'integer', minimum: 1, maximum: 100, default: 10 },
		sinceId: { type: 'string', format: 'misskey:id' },
		untilId: { type: 'string', format: 'misskey:id' },
		sinceDate: { type: 'integer' },
		untilDate: { type: 'integer' },
		allowPartial: { type: 'boolean', default: false }, // true is recommended but for compatibility false by default
		withFiles: { type: 'boolean', default: false },
	},
	required: ['userId'],
} as const;

@Injectable()
export default class extends Endpoint<typeof meta, typeof paramDef> { // eslint-disable-line import/no-default-export
	constructor(
		@Inject(DI.notesRepository)
		private notesRepository: NotesRepository,

		private noteEntityService: NoteEntityService,
		private queryService: QueryService,
		private cacheService: CacheService,
		private idService: IdService,
		private fanoutTimelineEndpointService: FanoutTimelineEndpointService,
		private metaService: MetaService,
	) {
		super(meta, paramDef, async (ps, me) => {
			const untilId = ps.untilId ?? (ps.untilDate ? this.idService.gen(ps.untilDate!) : null);
			const sinceId = ps.sinceId ?? (ps.sinceDate ? this.idService.gen(ps.sinceDate!) : null);
			const isSelf = me && (me.id === ps.userId);

			const serverSettings = await this.metaService.fetch();

			if (ps.withReplies && ps.withFiles) throw new ApiError(meta.errors.bothWithRepliesAndWithFiles);

			// early return if me is blocked by requesting user
			if (me != null) {
				const userIdsWhoBlockingMe = await this.cacheService.userBlockedCache.fetch(me.id);
				if (userIdsWhoBlockingMe.has(ps.userId)) {
					return [];
				}
			}

			if (!serverSettings.enableFanoutTimeline) {
				const timeline = await this.getFromDb({
					untilId,
					sinceId,
					limit: ps.limit,
					userId: ps.userId,
					withChannelNotes: ps.withChannelNotes,
					withFiles: ps.withFiles,
					withRenotes: ps.withRenotes,
					withQuotes: ps.withQuotes,
					withNonPublic: ps.withNonPublic,
					withRepliesToOthers: ps.withReplies,
					withRepliesToSelf: ps.withRepliesToSelf,
				}, me);

				return await this.noteEntityService.packMany(timeline, me);
			}

			const redisTimelines: FanoutTimelineName[] = [ps.withFiles ? `userTimelineWithFiles:${ps.userId}` : `userTimeline:${ps.userId}`];

			if (ps.withReplies) redisTimelines.push(`userTimelineWithReplies:${ps.userId}`);
			if (ps.withChannelNotes) redisTimelines.push(`userTimelineWithChannel:${ps.userId}`);

			const isFollowing = me && Object.hasOwn(await this.cacheService.userFollowingsCache.fetch(me.id), ps.userId);

			const timeline = await this.fanoutTimelineEndpointService.timeline({
				untilId,
				sinceId,
				limit: ps.limit,
				allowPartial: ps.allowPartial,
				me,
				redisTimelines,
				useDbFallback: true,
				ignoreAuthorFromMute: true,
				excludeReplies: ps.withChannelNotes && !ps.withReplies, // userTimelineWithChannel may include replies
				excludeNoFiles: ps.withChannelNotes && ps.withFiles, // userTimelineWithChannel may include notes without files
				excludePureRenotes: !ps.withRenotes,
				noteFilter: note => {
					if (note.channel?.isSensitive && !isSelf) return false;
					if (note.visibility === 'specified' && (!me || (me.id !== note.userId && !note.visibleUserIds.some(v => v === me.id)))) return false;
					if (note.visibility === 'followers' && !isFollowing && !isSelf) return false;

					// These are handled by DB fallback, but we duplicate them here in case a timeline was already populated with notes
					if (!ps.withRepliesToSelf && note.reply?.userId === note.userId) return false;
					if (!ps.withQuotes && isRenote(note) && isQuote(note)) return false;
					if (!ps.withNonPublic && note.visibility !== 'public') return false;

					return true;
				},
				dbFallback: async (untilId, sinceId, limit) => await this.getFromDb({
					untilId,
					sinceId,
					limit,
					userId: ps.userId,
					withChannelNotes: ps.withChannelNotes,
					withFiles: ps.withFiles,
					withRenotes: ps.withRenotes,
					withQuotes: ps.withQuotes,
					withNonPublic: ps.withNonPublic,
					withRepliesToOthers: ps.withReplies,
					withRepliesToSelf: ps.withRepliesToSelf,
				}, me),
			});

			return timeline;
		});
	}

	private async getFromDb(ps: {
		untilId: string | null,
		sinceId: string | null,
		limit: number,
		userId: string,
		withChannelNotes: boolean,
		withFiles: boolean,
		withRenotes: boolean,
		withQuotes: boolean,
		withNonPublic: boolean,
		withRepliesToOthers: boolean,
		withRepliesToSelf: boolean,
	}, me: MiLocalUser | null) {
		const isSelf = me && (me.id === ps.userId);

		const query = this.queryService.makePaginationQuery(this.notesRepository.createQueryBuilder('note'), ps.sinceId, ps.untilId)
			.andWhere('note.userId = :userId', { userId: ps.userId })
			.innerJoinAndSelect('note.user', 'user')
			.leftJoinAndSelect('note.reply', 'reply')
			.leftJoinAndSelect('note.renote', 'renote')
			.leftJoinAndSelect('note.channel', 'channel')
			.leftJoinAndSelect('reply.user', 'replyUser')
			.leftJoinAndSelect('renote.user', 'renoteUser');

		if (ps.withChannelNotes) {
			if (!isSelf) query.andWhere(new Brackets(qb => {
				qb.orWhere('note.channelId IS NULL');
				qb.orWhere('channel.isSensitive = false');
			}));
		} else {
			query.andWhere('note.channelId IS NULL');
		}

		this.queryService.generateVisibilityQuery(query, me);
		if (me) {
			this.queryService.generateMutedUserQuery(query, me, { id: ps.userId });
			this.queryService.generateBlockedUserQuery(query, me);
		}

		if (ps.withFiles) {
			query.andWhere('note.fileIds != \'{}\'');
		}

		if (!ps.withRenotes && !ps.withQuotes) {
			query.andWhere('note.renoteId IS NULL');
		} else if (!ps.withRenotes) {
			query.andWhere(new Brackets(qb => {
				qb.orWhere('note.userId != :userId', { userId: ps.userId });
				qb.orWhere('note.renoteId IS NULL');
				qb.orWhere('note.text IS NOT NULL');
				qb.orWhere('note.fileIds != \'{}\'');
				qb.orWhere('0 < (SELECT COUNT(*) FROM poll WHERE poll."noteId" = note.id)');
			}));
		} else if (!ps.withQuotes) {
			query.andWhere(`
				(
					note."renoteId" IS NULL
					OR (
						note.text IS NULL
						AND note.cw IS NULL
						AND note."replyId" IS NULL
						AND note."hasPoll" IS FALSE
						AND note."fileIds" = '{}'
					)
				)
			`);
		}

		if (!ps.withRepliesToOthers && !ps.withRepliesToSelf) {
			query.andWhere('reply.id IS NULL');
		} else if (!ps.withRepliesToOthers) {
			query.andWhere('(reply.id IS NULL OR reply."userId" = note."userId")');
		} else if (!ps.withRepliesToSelf) {
			query.andWhere('(reply.id IS NULL OR reply."userId" != note."userId")');
		}

		if (!ps.withNonPublic) {
			query.andWhere('note.visibility = \'public\'');
		}

		return await query.limit(ps.limit).getMany();
	}
}
